package cz.agents.alite.jbullet;

import com.bulletphysics.collision.dispatch.CollisionObject;
import com.bulletphysics.collision.shapes.BoxShape;
import com.bulletphysics.collision.shapes.CollisionShape;
import com.bulletphysics.collision.shapes.CompoundShape;
import com.bulletphysics.dynamics.RigidBody;
import com.bulletphysics.dynamics.vehicle.*;
import com.bulletphysics.linearmath.Transform;
import javax.vecmath.Quat4f;
import javax.vecmath.Vector3f;

public class RegularCarModel implements Cloneable {

    public static final CarDefinition carDefinition = new TruckCarDefinition();
    private VehicleRaycaster vehicleRayCaster;
    private RaycastVehicle vehicle;
    private RigidBody carChassis;
    private CollisionShape chassisShape;
    private CompoundShape compound;
    private float vehicleSteering;
    private float engineForce;
    private float breakingForce;

    public RegularCarModel(Vector3f initialTranslation) {
        init(initialTranslation);
    }

    private void init(Vector3f initialTranslation) {
        vehicleRayCaster = new DefaultVehicleRaycaster(BulletPhysics.getInstance().getDynamicsWorld());
        chassisShape = new BoxShape(new Vector3f(
                0.5f * carDefinition.getCollisionBox().x,
                0.5f * carDefinition.getCollisionBox().z,
                0.5f * carDefinition.getCollisionBox().y));
        compound = new CompoundShape();

        Transform localTrans = new Transform();
        localTrans.setIdentity();
        localTrans.origin.set(0, 1, 0);
        compound.addChildShape(localTrans, chassisShape);

        Transform transform = new Transform();
        transform.setIdentity();
        transform.origin.set(initialTranslation);
        carChassis = BulletPhysics.getInstance().createRigidBodyAndAddObject(carDefinition.getMass(), compound, transform);

        carChassis.setActivationState(CollisionObject.DISABLE_DEACTIVATION);
        vehicle = new RaycastVehicle(carDefinition.getTuning(), carChassis, vehicleRayCaster);
        BulletPhysics.getInstance().addVehicle(vehicle);
        vehicle.setCoordinateSystem(0, 1, 2);

        // add wheels
        for (WheelDefinition wheelDefinition : carDefinition.getWheelDefinitions()) {
            boolean isSteered = wheelDefinition.isSteered;
            Vector3f connectionPoint = new Vector3f(wheelDefinition.offset.x, wheelDefinition.offset.z, -wheelDefinition.offset.y);
            getVehicle().addWheel(
                    connectionPoint,
                    carDefinition.getWheelDirection(),
                    carDefinition.getWheelAxle(),
                    carDefinition.getSuspensionRestLength(),
                    wheelDefinition.radius,
                    carDefinition.getTuning(),
                    isSteered);
        }

        for (int i = 0; i < getVehicle().getNumWheels(); i++) {
            WheelInfo wheel = getVehicle().getWheelInfo(i);
            wheel.suspensionStiffness = carDefinition.getSuspensionStiffness();
            wheel.wheelsDampingRelaxation = carDefinition.getSuspensionDamping();
            wheel.wheelsDampingCompression = carDefinition.getSuspensionCompression();
            wheel.frictionSlip = carDefinition.getWheelFriction();
            wheel.rollInfluence = carDefinition.getWheelRollInfluence();
        }
    }

    public void move() {
        int wheelIndex = 0;
        for (WheelDefinition wheelDefinition : carDefinition.getWheelDefinitions()) {
            if (wheelDefinition.isSteered) {
                vehicle.setSteeringValue(vehicleSteering, wheelIndex);
            }
            if (wheelDefinition.isPropulsive) {
                vehicle.applyEngineForce(engineForce, wheelIndex);
            }
            if (wheelDefinition.isBreaking) {
                vehicle.setBrake(breakingForce, wheelIndex);
            }

            ++wheelIndex;
        }
    }

    public Vector3f getPosition() {
        Transform tr = new Transform();
        carChassis.getWorldTransform(tr);
        return tr.origin;
    }

    public Quat4f getRotation() {
        Transform tr = new Transform();
        carChassis.getWorldTransform(tr);
        return Converter.convert(tr.basis, new Quat4f());
    }

    public Vector3f getForwardVector() {
        Vector3f ret = new Vector3f();
        vehicle.getForwardVector(ret);
        return ret;
    }

    public float getSpeed() {
        return vehicle.getCurrentSpeedKmHour();
    }

    private Quat4f getWheelRotation(int wheelIndex) {
        Transform wheelWorldTr = new Transform();
        vehicle.updateWheelTransform(wheelIndex);
        vehicle.getWheelTransformWS(wheelIndex, wheelWorldTr);
        Quat4f wheelWorldRot = Converter.convert(wheelWorldTr.basis, new Quat4f());
        Quat4f parentWorldRot = getRotation();
        parentWorldRot.inverse();
        parentWorldRot.mul(wheelWorldRot);
        return parentWorldRot;
    }

    private Vector3f getWheelTranslation(int wheelIndex) {;
        Transform wheelWorldTr = new Transform();
        vehicle.updateWheelTransform(wheelIndex);
        vehicle.getWheelTransformWS(wheelIndex, wheelWorldTr);

        Transform tr = new Transform();
        carChassis.getWorldTransform(tr);

        tr.inverse();
        tr.mul(wheelWorldTr);

        return tr.origin;
    }

    public Quat4f getWheelRotation(String wheelName) {
        return getWheelRotation(translateWheelNameToIndex(wheelName));
    }

    public Vector3f getWheelTranslation(String wheelName) {
        return getWheelTranslation(translateWheelNameToIndex(wheelName));
    }

    private int translateWheelNameToIndex(String wheelName) {
        if (wheelName.equals("LF")) {
            return 0;
        } else if (wheelName.equals("RF")) {
            return 1;
        } else if (wheelName.equals("LB")) {
            return 2;
        } else if (wheelName.equals("RB")) {
            return 3;
        } else if (wheelName.equals("LBB")) {
            return 4;
        } else if (wheelName.equals("RBB")) {
            return 5;
        } else {
            throw new RuntimeException("Unknown wheel name: " + wheelName);
        }
    }

    public CollisionShape getCollisionShape() {
        return chassisShape;
    }

    public CompoundShape getCompound() {
        return compound;
    }

    public RigidBody getRigidBody() {
        return carChassis;
    }

    public void setSteering(float value) {
        vehicleSteering = value;
    }

    public void setEngineForce(float engineForce) {
        this.engineForce = engineForce;
    }

    public void setBreakingForce(float breakingForce) {
        this.breakingForce = breakingForce;
    }

    public RaycastVehicle getVehicle() {
        return vehicle;
    }

    @Override
    public RegularCarModel clone() {
        RegularCarModel carModel;
        try {
            carModel = (RegularCarModel) super.clone();
            carModel.init(getPosition());
            return carModel;
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        return null;
    }

    public static class MdarsCarDefinition extends CarDefinition {

        public MdarsCarDefinition() {
            collisionBox = new Vector3f(1.6f, 2.5f, 1.2f);
            wheelDefinitions = new WheelDefinition[]{
                new WheelDefinition(0.4f, new Vector3f(0.76f, -0.9f, 1.0f), true, true, true),
                new WheelDefinition(0.4f, new Vector3f(-0.76f, -0.9f, 1.0f), true, true, true),
                new WheelDefinition(0.4f, new Vector3f(-0.76f, 0.9f, 1.0f), false, false, true),
                new WheelDefinition(0.4f, new Vector3f(0.76f, 0.9f, 1.0f), false, false, true),};
            mass = 500;

            wheelDirection = new Vector3f(0, -1, 0);
            wheelAxle = new Vector3f(-1, 0, 0);
            wheelFriction = 1000f;
            wheelRollInfluence = 0.1f;
            suspensionStiffness = 20.f;
            suspensionDamping = 2.3f;
            suspensionCompression = 4.4f;
            suspensionRestLength = 0.6f;
            tuning = new VehicleTuning();
        }
    }

    public static class TruckCarDefinition extends CarDefinition {

        public TruckCarDefinition() {
            collisionBox = new Vector3f(1.8f, 5.5f, 1.5f);
            wheelDefinitions = new WheelDefinition[]{
                new WheelDefinition(0.5f, new Vector3f(0.756f, -1.855f, 1.2f), true, true, true),
                new WheelDefinition(0.5f, new Vector3f(-0.756f, -1.855f, 1.2f), true, true, true),
                new WheelDefinition(0.5f, new Vector3f(-0.756f, 1.29f, 1.2f), false, true, true),
                new WheelDefinition(0.5f, new Vector3f(0.756f, 1.29f, 1.2f), false, true, true),
                new WheelDefinition(0.5f, new Vector3f(-0.756f, 2.29f, 1.2f), false, true, true),
                new WheelDefinition(0.5f, new Vector3f(0.756f, 2.29f, 1.2f), false, true, true),};
            mass = 500;

            wheelDirection = new Vector3f(0, -1, 0);
            wheelAxle = new Vector3f(-1, 0, 0);
            wheelFriction = 1000f;
            wheelRollInfluence = 0.1f;
            suspensionStiffness = 20.f;
            suspensionDamping = 2.3f;
            suspensionCompression = 4.4f;
            suspensionRestLength = 0.6f;
            tuning = new VehicleTuning();
        }
    }

    public abstract static class CarDefinition {

        protected Vector3f collisionBox;
        protected WheelDefinition[] wheelDefinitions;
        protected float mass;
        protected Vector3f wheelDirection;
        protected Vector3f wheelAxle;
        protected float wheelFriction;
        protected float wheelRollInfluence;
        protected float suspensionStiffness;
        protected float suspensionDamping;
        protected float suspensionCompression;
        protected float suspensionRestLength;
        protected VehicleTuning tuning;

        public Vector3f getCollisionBox() {
            return collisionBox;
        }

        public WheelDefinition[] getWheelDefinitions() {
            return wheelDefinitions;
        }

        public float getMass() {
            return mass;
        }

        public Vector3f getWheelDirection() {
            return wheelDirection;
        }

        public Vector3f getWheelAxle() {
            return wheelAxle;
        }

        public float getWheelFriction() {
            return wheelFriction;
        }

        public float getWheelRollInfluence() {
            return wheelRollInfluence;
        }

        public float getSuspensionStiffness() {
            return suspensionStiffness;
        }

        public float getSuspensionDamping() {
            return suspensionDamping;
        }

        public float getSuspensionCompression() {
            return suspensionCompression;
        }

        public float getSuspensionRestLength() {
            return suspensionRestLength;
        }

        public VehicleTuning getTuning() {
            return tuning;
        }
    }

    public static class WheelDefinition {

        public final float radius;
        public final Vector3f offset;
        public final boolean isSteered;
        public final boolean isPropulsive;
        public final boolean isBreaking;

        public WheelDefinition(float radius, Vector3f offset, boolean isSteered, boolean isPropulsive, boolean isBreaking) {
            this.radius = radius;
            this.offset = offset;
            this.isSteered = isSteered;
            this.isPropulsive = isPropulsive;
            this.isBreaking = isBreaking;
        }
    }
}
