package tacv.universe.entity.embodiment;

import com.jme3.math.Quaternion;
import com.jme3.math.Vector3f;
import cz.agents.alite.jbullet.BulletPhysics;
import cz.agents.alite.jbullet.RegularCarModel;
import tacv.settings.CoordsSettings;
import tacv.universe.entity.PositionedEntity;
import tacv.util.Converter;
import tacv.util.Point;
import tacv.vis.VisualPlanLayer.VisualPlan;

public class AutomaticRegularCar implements PositionedEntity {

    private static final double MAX_ENGINE_FORCE = 2000.00f;
    private static final double MAX_BREAKING_FORCE = 100.00f;
    private static final double MAX_STEERING = 1.00f;
    private static final double BREAKING_DISTANCE = 4.0;
    private static final double STEERING_SPEED = 35.0;
    private RegularCarModel carModel;
    private CarRegulator carRegulator;
    private String name;
    private Point waypoint;
    private Float targetVelocity;
    private boolean isForward;
    private VisualPlan visualPlan;

    public AutomaticRegularCar(String name, Point actualPosition) {
        this.name = name;
        javax.vecmath.Vector3f initialTranslation = Converter.convertVec(actualPosition, CoordsSettings.PROJECT_COORDS, new javax.vecmath.Vector3f(), CoordsSettings.BULLET_COORDS);
        this.carModel = new RegularCarModel(initialTranslation);
        this.carRegulator = new CarRegulator();
    }

    public void move(long duration, long currentTime) {
        BulletPhysics.getInstance().stepSimulation(currentTime);
        carModel.move();

        // by default stop
        float vehicleSteering = 0;
        float engineForce = 0;
        float breakingForce = (float) MAX_BREAKING_FORCE;

        // if target exists go towards it
        if (waypoint != null) {
            javax.vecmath.Vector3f bulletWaypoint = Converter.convertVec(waypoint, CoordsSettings.PROJECT_COORDS, new javax.vecmath.Vector3f(), CoordsSettings.BULLET_COORDS);
            javax.vecmath.Vector3f bulletPosition = carModel.getPosition();

            // count desired heading
            javax.vecmath.Vector3f bulletDesiredHeading = (javax.vecmath.Vector3f) bulletWaypoint.clone();
            bulletDesiredHeading.sub(bulletPosition);
            if (!isForward) {
                // for backward movement flip heading
                bulletDesiredHeading.scale(-1);
            }

            // count desired speed
            double desiredSpeed = targetVelocity;

            // regulate
            double speed = getSpeed();
            
            if (!isForward) {
                speed = -speed;
            }
            
            javax.vecmath.Vector3f bulletHeading = carModel.getForwardVector();
            CarRegulator.Action action = carRegulator.calculateAction(desiredSpeed, bulletDesiredHeading, speed, bulletHeading);

            // control
            vehicleSteering = toInterval(action.steer, -MAX_STEERING, MAX_STEERING);
            engineForce = toInterval(action.force, 0, MAX_ENGINE_FORCE);
            breakingForce = toInterval(-action.force, 0, MAX_BREAKING_FORCE);

            if (!isForward) {
                // for backward movement invert steering and engine force
                vehicleSteering = -vehicleSteering;
                engineForce = -engineForce;
            }

        }

        carModel.setSteering(vehicleSteering);
        carModel.setEngineForce(engineForce);
        carModel.setBreakingForce(breakingForce);
    }

    public Double getDistanceToWaypoint() {
        if (waypoint == null) {
            return null;
        }

        Vector3f carPosition = Converter.convertVec(carModel.getPosition(), CoordsSettings.BULLET_COORDS, new Vector3f(), CoordsSettings.PROJECT_COORDS);
        double sqrt = Math.sqrt(Math.pow(waypoint.x - carPosition.x, 2) + Math.pow(waypoint.y - carPosition.y, 2));
        return sqrt;
    }

    public RegularCarModel getCarModel() {
        return carModel;
    }

    public String getName() {
        return name;
    }

    public Point getWaypoint() {
        return waypoint;
    }

    public void setWaypoint(Point waypoint, Float targetVelocity, boolean isForward) {
        this.waypoint = waypoint;
        this.targetVelocity = targetVelocity;
        this.isForward = isForward;
    }

    public Point getActualPosition() {
        return Converter.convertVec(carModel.getPosition(), CoordsSettings.BULLET_COORDS, new Point(), CoordsSettings.PROJECT_COORDS);
    }

    @Override
    public Point getPosition() {
        return getActualPosition();
    }

    public Vector3f getTranslation() {
        return Converter.convertVec(carModel.getPosition(), CoordsSettings.BULLET_COORDS, new Vector3f(), CoordsSettings.PROJECT_COORDS);
    }

    public Vector3f getVelocity() {
        return Converter.convertVec(carModel.getRigidBody().getLinearVelocity(new javax.vecmath.Vector3f()), CoordsSettings.BULLET_COORDS, new Vector3f(), CoordsSettings.PROJECT_COORDS);
    }

    public float getSpeed() {
        return carModel.getSpeed();
    }

    public Quaternion getWheelRotation(String wheelName) {
        return Converter.convertQuat(carModel.getWheelRotation(wheelName), CoordsSettings.BULLET_COORDS, new Quaternion(), CoordsSettings.PROJECT_COORDS);
    }

    public Vector3f getWheelTranslation(String wheelName) {
        return Converter.convertVec(carModel.getWheelTranslation(wheelName), CoordsSettings.BULLET_COORDS, new Vector3f(), CoordsSettings.PROJECT_COORDS);
    }

    public Vector3f getHeading() {
        return Converter.convertVec(carModel.getForwardVector(), CoordsSettings.BULLET_COORDS, new Vector3f(), CoordsSettings.PROJECT_COORDS);
    }

    public Quaternion getRotation() {
        return Converter.convertQuat(carModel.getRotation(), CoordsSettings.BULLET_COORDS, new Quaternion(), CoordsSettings.PROJECT_COORDS);
    }

    public VisualPlan getVisualPlan() {
        return visualPlan;
    }

    public void setVisualPlan(VisualPlan visualPlan) {
        this.visualPlan = visualPlan;
    }

    @Override
    public String toString() {
        return "v: " + getSpeed();
    }

    private float toInterval(double val, double min, double max) {
        return (float) Math.min(Math.max(val, min), max);
    }
}
