package tacv.universe.world;

import com.bulletphysics.collision.shapes.BvhTriangleMeshShape;
import com.bulletphysics.collision.shapes.TriangleCallback;
import com.bulletphysics.linearmath.Transform;
import cz.agents.alite.jbullet.BulletPhysics;
import java.util.ArrayList;
import java.util.List;
import javax.vecmath.Vector3f;
import tacv.universe.world.xml.CollisionMesh;
import tacv.util.XmlLoader;

public class WorldInfo {

    private static final double SAFETY_RANGE = 1.2;
    private static final String ONTOLOGY_PACKAGE = "tacv.universe.world.xml";
    private static final String XML_FILE_PATH = "resources/map/collisionmesh.xml";
    //xml load
    private CollisionMesh xmlCollision;
    //terrain
    private Terrain terrain;
    // collision zones
    private CollisionZone collisionZone;
    //bullet things
    private List<BvhTriangleMeshShape> collisionMeshes;
    private BvhTriangleMeshShape groundMeshShape;
    private Vector3f worldMin;
    private Vector3f worldMax;


    public void initTerrain() {
        checkTerrain();
    }

    public void initBulletPhysics() {
        checkXMLLoad();
        checkCollisionMeshes();
        BulletPhysics.init(worldMin, worldMax);
        feedBulletPhysiscs();
    }

    private void checkXMLLoad() {
        if (xmlCollision == null) {
            this.xmlCollision = (CollisionMesh) XmlLoader.loadXml(XML_FILE_PATH, ONTOLOGY_PACKAGE);
            float maxX = (float) xmlCollision.getHeightmap().getTerrainSizeX();
            float maxY = (float) xmlCollision.getHeightmap().getTerrainSizeY();
            float maxZ = Math.max(maxX, maxY);

            this.worldMax = new Vector3f(maxX, maxY, maxZ);
            this.worldMin = new Vector3f(-maxX, -maxY, -maxZ);
        }
    }

    private void checkTerrain() {
        checkXMLLoad();
        if (terrain == null) {
            this.terrain = Terrain.create(xmlCollision);
        }
    }

    private void checkCollisionZone() {
        checkXMLLoad();
        if (collisionZone == null) {
            this.collisionZone = new CollisionZone(CollisionZoneLoader.loadZones(xmlCollision, SAFETY_RANGE));
        }
    }

    public Terrain getTerrain() {
        checkTerrain();
        return terrain;
    }

    public CollisionZone getCollisionZone() {
        checkCollisionZone();
        return collisionZone;
    }

    private void checkCollisionMeshes() {
        if (collisionMeshes == null) {
            collisionMeshes = CollisionMeshLoader.loadCollisionMesh(xmlCollision);
        }
    }

    private void feedBulletPhysiscs() {
        Transform groundTransform = new Transform();
        groundTransform.setIdentity();
        groundMeshShape = collisionMeshes.get(0);

        for (BvhTriangleMeshShape bvhTriangleMeshShape : collisionMeshes) {
            BulletPhysics.getInstance().createRigidBodyAndAddObject(0, bvhTriangleMeshShape, groundTransform);
        }
    }

    public float getAltitude(float x, float y) {
        checkTerrain();
        return terrain.getAltitude(x, y);
    }

    public float getEstimatedAltitude(float x, float y) {
        // how to simply return a variable from a callback ;)
        final List<Float> height = new ArrayList<Float>();
        groundMeshShape.performRaycast(new TriangleCallback() {

            @Override
            public void processTriangle(Vector3f[] triangle, int partId, int triangleIndex) {
                // not precise, but simple
                Float f = triangle[0].y + triangle[1].y + triangle[2].y;
                f = f / 3.0f;
                height.add(f);
            }
        }, new Vector3f(x,worldMax.z, -y), new Vector3f(x, worldMin.z, -y));

        return height.get(0);
    }
}
